// import pkg_silo::*;

module flash_driver #(
    parameter                           BC_MSB = 9,          //byte count MSB
    parameter                           NUM_PORTS = 4
)(
    input                               clk,
    flash_port                          ports [0:3],
    
    // Flash pins
    inout               [3:0]           IO,
    output                              CS
);
    logic [23:0] addr_to_flash = '0;
    logic [BC_MSB:0] bytes_to_grab = '0;
    logic flash_phy_enable = '0;
    wire [7:0] byte_from_flash;
    wire we_from_flash;
    logic [BC_MSB:0] byte_count = '0;
    wire last_count = (byte_count==bytes_to_grab);
    
    logic [23:0]    addr_start      [0:NUM_PORTS-1];
    logic [9:0]     length_bytes    [0:NUM_PORTS-1];
    logic           request         [0:NUM_PORTS-1];
    logic [7:0]     data_byte       [0:NUM_PORTS-1];
    logic           byte_valid      [0:NUM_PORTS-1];
    logic           done            [0:NUM_PORTS-1];
    logic           running         [0:NUM_PORTS-1];
    
    genvar n;
    generate
        for(n=0; n<NUM_PORTS; n++) begin: port_assignments
            assign addr_start[n]        = ports[n].addr_start;
            assign length_bytes[n]      = ports[n].length_bytes;
            assign request[n]           = ports[n].request;
            assign ports[n].data_byte   = data_byte[n];
            assign ports[n].byte_valid  = byte_valid[n];
            assign ports[n].done        = done[n];
            assign ports[n].running     = running[n];
        end
    endgenerate

    enum reg [3:0] {
        IDLE,
        LOAD,
        TRANSFER
    } state, nx_state;
    
    logic [3:0] port = '0;
    
    initial begin
        state = IDLE;
        for(int k=0; k<NUM_PORTS; k++) begin
            done[k] = '1;
            byte_valid[k] = '0;
            data_byte[k] = '0;
            running[k] = '0;
        end
    end
    
    function any_port_request;
        automatic logic result = '0;
        for(int k=0; k<NUM_PORTS; k++) result |= request[k];
        return result;
    endfunction
    
    function [3:0] lowest_port_request;
        automatic logic [3:0] result = '0;
        for(int k=0; k<NUM_PORTS; k++) begin
            if(request[k]) begin
                result = k[3:0];
                break;
            end
        end
        return result;
    endfunction
    
    always_comb begin
        nx_state = IDLE;
        case(state)
            IDLE:           if(any_port_request)        nx_state = LOAD;
                            else                        nx_state = IDLE;
            LOAD:                                       nx_state = TRANSFER;
            TRANSFER:       if(last_count)              nx_state = IDLE;
                            else                        nx_state = TRANSFER;
        endcase
    end
    
    always @(posedge clk) begin
        state <= nx_state;
        
        if(state==LOAD)                                             running[port]   <= '1;
        else if(nx_state==IDLE)     for(int k=0; k<NUM_PORTS; k++)  running[k]      <= '0;
        
        for(int k=0; k<NUM_PORTS; k++) begin
            if(request[k])                                          done[k] <= '0;
            else if((state==TRANSFER)&(nx_state==IDLE)&(port==k))   done[k] <= '1;
        end
        
        if(nx_state==LOAD) begin
            port <= lowest_port_request;
            addr_to_flash <= addr_start[lowest_port_request];
            bytes_to_grab <= length_bytes[lowest_port_request];
            flash_phy_enable <= '1;
        end else if(nx_state==IDLE) begin
            flash_phy_enable <= '0;
        end
        
        if(state==IDLE)                         byte_count <= '0;
        else if(we_from_flash)                  byte_count <= byte_count + 1'd1;
        
        if((state==TRANSFER) & we_from_flash)   byte_valid[port] <= '1;
        else for(int k=0; k<NUM_PORTS; k++)     byte_valid[k]    <= '0;
        
        if(we_from_flash) data_byte[port] <= byte_from_flash;
    end
    
    phy_W25Q16CV m_phy(
        .IO,
        .CLK            (clk),
        .CS,
        .address        (addr_to_flash),
        .enable         (flash_phy_enable),
        .byte_out       (byte_from_flash),
        .byte_out_we    (we_from_flash)
    );
endmodule
