module phy_W25Q16CV(
    inout       [3:0]   IO,
    input               CLK,
    input       [23:0]  address,
    input               enable,
    output  reg         CS,
    output  reg [7:0]   byte_out,
    output  reg         byte_out_we
);

    localparam MBITS = 8'b0010_0000;    //For continuous read mode

    reg [3:0] bit_count;
    wire nibble_count = bit_count[0];   //0 for first nibble, 1 for second nibble

    reg [7:0] shift_out [0:3];
    reg [3:0] shift_in = '0;
    reg [3:0] shift_in_d1 = '0;
    reg is_out = '0;

    assign IO[0] = is_out ? shift_out[0][7] : 'Z;
    assign IO[1] = is_out ? shift_out[1][7] : 'Z;
    assign IO[2] = is_out ? shift_out[2][7] : 'Z;
    assign IO[3] = is_out ? shift_out[3][7] : 'Z;

    enum reg [3:0] {
        IDLE,               //Startup state
        WAIT,               //True idle state
        ADDRESS,            //Address and MBITS = 8 clocks
        DUMMY,              //Dummy bits = 2 clocks
        DATA                //Data = 2 clocks per byte
    } state, nx_state;

    initial begin
        state = IDLE;
        CS = '1;
    end

    wire load_address       = (state==WAIT) & (nx_state==ADDRESS);
    wire make_input         = (nx_state==DUMMY);
    wire cs_assert          = (nx_state==ADDRESS);
    wire cs_clear           = (nx_state==WAIT);

    wire time_to_write      = (state==DATA) & nibble_count;
    logic time_to_write_d1  = '0;

    // wire time_to_write      = ((state==DATA) | (nx_state==DATA)) & nibble_count;
    //The nx_state case was added to catch the very first byte

    always @(*) begin
        nx_state = IDLE;
        case(state)
            IDLE:                                                   nx_state = WAIT;
            WAIT:                       if(enable)                  nx_state = ADDRESS;
                                        else                        nx_state = WAIT;
            ADDRESS:                    if(enable)
                                            if(bit_count==4'd7)     nx_state = DUMMY;
                                            else                    nx_state = ADDRESS;
                                        else                        nx_state = WAIT;
            DUMMY:                      if(enable)
                                            if(bit_count==4'd3)     nx_state = DATA;
                                            else                    nx_state = DUMMY;
                                        else                        nx_state = WAIT;
            DATA:                       if(enable)                  nx_state = DATA;
                                        else                        nx_state = WAIT;
        endcase
    end

    always @(negedge CLK) begin
        state <= nx_state;
        
        if(state!=nx_state)     bit_count <= '0;
        else if(state==IDLE)    bit_count <= '0;
        else                    bit_count <= bit_count + 1'd1;
        
        if(cs_assert)       CS <= '0;
        else if(cs_clear)   CS <= '1;
        
        if(load_address) begin
            is_out <= '1;
            shift_out[0] <= {address[20], address[16], address[12], address[ 8], address[4], address[0], MBITS[4], MBITS[0]};
            shift_out[1] <= {address[21], address[17], address[13], address[ 9], address[5], address[1], MBITS[5], MBITS[1]};
            shift_out[2] <= {address[22], address[18], address[14], address[10], address[6], address[2], MBITS[6], MBITS[2]};
            shift_out[3] <= {address[23], address[19], address[15], address[11], address[7], address[3], MBITS[7], MBITS[3]};
        end else if(make_input) begin
            is_out <= '0;
        end else begin
            shift_out[0] <= (shift_out[0]<<1);
            shift_out[1] <= (shift_out[1]<<1);
            shift_out[2] <= (shift_out[2]<<1);
            shift_out[3] <= (shift_out[3]<<1);
        end
        
        shift_in <= IO;
        shift_in_d1 <= shift_in;
        time_to_write_d1 <= time_to_write;
        if(time_to_write_d1) begin
            byte_out <= {shift_in_d1, shift_in};
            byte_out_we <= '1;
        end else begin
            byte_out_we <= '0;
        end
    end
endmodule
